#!/usr/bin/python3
# vim: set filetype=python
import os
import sys
import re
import json
import subprocess
import time

class EventFileOpen(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "open"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_openat", self.syscalls_sys_enter_openat)
        self.em.event_watch("syscalls/sys_exit_openat", self.syscalls_sys_exit_openat)
        self.em.event_watch("syscalls/sys_enter_open", self.syscalls_sys_enter_open)
        self.em.event_watch("syscalls/sys_exit_open", self.syscalls_sys_exit_open)

    def syscalls_sys_enter_openat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_openat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            if 'filename' in event['parameters']:
                filename = event['parameters']['filename']
            else:
                filename = "<none>"
            mode = event['parameters']['mode']
            flags = event['parameters']['flags']
            detail = f"{filename} mode={mode} flags={flags} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

    def syscalls_sys_enter_open(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_open(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            filename = event['parameters']['filename']
            mode = event['parameters']['mode']
            flags = event['parameters']['flags']
            detail = f"{filename} mode={mode} flags={flags} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileUnlink(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "unlink"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_unlink", self.syscalls_sys_enter_unlink)
        self.em.event_watch("syscalls/sys_exit_unlink", self.syscalls_sys_exit_unlink)
        self.em.event_watch("syscalls/sys_enter_unlinkat", self.syscalls_sys_enter_unlink)
        self.em.event_watch("syscalls/sys_exit_unlinkat", self.syscalls_sys_exit_unlink)


    def syscalls_sys_enter_unlink(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_unlink(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            pathname = event['parameters']['pathname']
            detail = f"{pathname} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileTruncate(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "truncate"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_truncate", self.syscalls_sys_enter_truncate)
        self.em.event_watch("syscalls/sys_exit_truncate", self.syscalls_sys_exit_truncate)

    def syscalls_sys_enter_truncate(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_truncate(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            pathname = event['parameters']['path']
            length = event['parameters']['length']
            detail = f"{pathname} length={length} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileChown(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "chown"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_chown", self.syscalls_sys_enter_chown)
        self.em.event_watch("syscalls/sys_exit_chown", self.syscalls_sys_exit_chown)
        self.em.event_watch("syscalls/sys_enter_fchownat", self.syscalls_sys_enter_fchownat)
        self.em.event_watch("syscalls/sys_exit_fchownat", self.syscalls_sys_exit_fchownat)


    def syscalls_sys_enter_chown(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_chown(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            filename = event['parameters']['filename']
            uid = event['parameters']['user']
            gid = event['parameters']['group']
            detail = f"{filename} uid={uid} gid={gid} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

    def syscalls_sys_enter_fchownat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_fchownat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            filename = event['parameters']['filename']
            uid = event['parameters']['user']
            gid = event['parameters']['group']
            detail = f"{filename} uid={uid} gid={gid} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileChmod(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "chmod"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_chmod", self.syscalls_sys_enter_chmod)
        self.em.event_watch("syscalls/sys_exit_chmod", self.syscalls_sys_exit_chmod)
        self.em.event_watch("syscalls/sys_enter_fchmodat", self.syscalls_sys_enter_fchmodat)
        self.em.event_watch("syscalls/sys_exit_fchmodat", self.syscalls_sys_exit_fchmodat)

    def syscalls_sys_enter_chmod(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_chmod(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            filename = event['parameters']['filename']
            mode = event['parameters']['mode']
            detail = f"{filename} mode={mode} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

    def syscalls_sys_enter_fchmodat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_fchmodat(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            filename = event['parameters']['filename']
            mode = event['parameters']['mode']
            detail = f"{filename} mode={mode} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileRmdir(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "rmdir"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_rmdir", self.syscalls_sys_enter_rmdir)
        self.em.event_watch("syscalls/sys_exit_rmdir", self.syscalls_sys_exit_rmdir)

    def syscalls_sys_enter_rmdir(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_rmdir(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            pathname = event['parameters']['pathname']
            detail = f"{pathname} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileMkdir(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "mkdir"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_mkdir", self.syscalls_sys_enter_mkdir)
        self.em.event_watch("syscalls/sys_exit_mkdir", self.syscalls_sys_exit_mkdir)
        self.em.event_watch("syscalls/sys_enter_mkdirat", self.syscalls_sys_enter_mkdir)
        self.em.event_watch("syscalls/sys_exit_mkdirat", self.syscalls_sys_exit_mkdir)

    def syscalls_sys_enter_mkdir(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_mkdir(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            pathname = event['parameters']['pathname']
            mode = event['parameters']['mode']
            detail = f"{pathname} mode={mode} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)

class EventFileSymlink(object):
    def __init__(self, event_mgr):
        self.syscall_hash = {}
        self.event_name = "symlink"
        self.em = event_mgr
        self.em.event_register(self.event_name)
        self.em.event_watch("syscalls/sys_enter_symlink", self.syscalls_sys_enter_symlink)
        self.em.event_watch("syscalls/sys_exit_symlink", self.syscalls_sys_exit_symlink)
        self.em.event_watch("syscalls/sys_enter_symlinkat", self.syscalls_sys_enter_symlink)
        self.em.event_watch("syscalls/sys_exit_symlinkat", self.syscalls_sys_exit_symlink)

    def syscalls_sys_enter_symlink(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        self.syscall_hash[syscall_nr] = event

    def syscalls_sys_exit_symlink(self, event):
        syscall_nr = event['parameters']['__syscall_nr']
        ret = event['parameters']['ret']
        if syscall_nr in self.syscall_hash:
            event = self.syscall_hash[syscall_nr]
            oldname = event['parameters']['oldname']
            newname = event['parameters']['newname']
            detail = f"oldname={oldname} newname={newname} ret={ret}"
            args = {
               "name":self.event_name,
               "comm": event["comm"],
               "pid": event["pid"],
               "uid": event["uid"],
               "detail": detail,
               "ok": True if ret >= 0 else False
            }
            del self.syscall_hash[syscall_nr]
            self.em.event_send(args)


