#!/usr/bin/python3
# vim: set filetype=python
import os
import sys
import re
import json
import subprocess
import time
import random
import shutil
import logging
import traceback
import pkgutil
from .schema import EventSchema

class TracePoint(object):
    def __init__(self, pid_filter=[], uid_filter=[], ignore_self=True):
        self.event_handlers = {}
        self.set_logger()
        #dir_name = "event_trace_%d" % random.randint(0, 1000000) 
        dir_name = "event_trace"
        self.obj_dir = os.path.join("/var/tmp", dir_name)
        self.trace_event_elf = os.path.join(self.obj_dir, "trace_event")
        self.trace_event_c = os.path.join(self.obj_dir, "trace_event.c")
        self.trace_bpf_o = []
        self.schema_h = os.path.join(self.obj_dir, "event_schema.h")
        self.schema = EventSchema()
        self.proc = None
        self.pid_filter = pid_filter
        self.uid_filter = uid_filter
        self.ignore_self = ignore_self
        self.pid = os.getpid()
        self.init_obj_dir()

    def set_logger(self):
        self.logger_name = 'pswatch'
        self.logger = logging.getLogger(self.logger_name)

    def init_obj_dir(self):
        if not os.path.exists(self.obj_dir):
            os.makedirs(self.obj_dir)

    def add_event_watch(self, event, handler):
        if event in self.event_handlers and self.event_handlers[event] != None:
            self.event_handlers[event].append(handler)
        else:
            self.event_handlers[event] = [handler]

    def delete_event_watch(self, event, handler):
        pass

    def start(self):
        pass

    def get_event_list(self):
        events = []
        for event in self.event_handlers.keys():
            events.append(event)
        return events

    def get_event_id(self, event):
        event_id = event.replace('/', '_')
        event_id = event_id.lower()
        return event_id

    def build_bpf_c(self, event):
        event_id = event.replace('/', '_')
        event_id = event_id.lower()
        bpf_c_file = os.path.join(self.obj_dir, f"{event_id}.bpf.c")
        event_func = f"func_{event_id}"
        event_schema = self.schema.get_event_schema_name(event)
        content="""#include "trace_event.bpf.c"
EVENT_TRACE_FUNC("tracepoint/%s", %s, %s)
""" % (event, event_func, event_schema)
        with open(bpf_c_file, 'w') as fp:
            fp.write(content)
        return bpf_c_file

    def build_bpf_o(self, event):
        event_id = event.replace('/', '_')
        event_id = event_id.lower()
        bpf_o = os.path.join(self.obj_dir, f"{event_id}.bpf.o")
        bpf_c = self.build_bpf_c(event)
        cmd = f"clang -O2 -target bpf -c {bpf_c} -o {bpf_o}"
        os.popen(cmd)
        self.trace_bpf_o.append(bpf_o)

    def copy_from_pkg(self, src, dst):
        data = pkgutil.get_data(__package__, src)
        with open(dst, 'wb') as fd:
            fd.write(data)

    def build_trace_objs(self):
        events = self.get_event_list()
        self.schema.build(events, self.schema_h)
        dst_file = os.path.join(self.obj_dir, 'trace_event.bpf.c')
        self.copy_from_pkg('trace_event/trace_event.bpf.c', dst_file)
        dst_file = os.path.join(self.obj_dir, 'event.h')
        self.copy_from_pkg('trace_event/event.h', dst_file)
        dst_file = os.path.join(self.obj_dir, 'trace_event.c')
        self.copy_from_pkg('trace_event/trace_event.c', dst_file)
        self.trace_bpf_o = []
        for event in events:
            self.build_bpf_o(event)
        cmd = f"gcc {self.trace_event_c} -g -L/usr/lib64/ -l:libbpf.a -ljson-c -lelf -lz -lpthread -o {self.trace_event_elf}"
        os.popen(cmd)

    def call_event_handlers(self, event_obj):
        event = event_obj['event']
        if self.ignore_self:
            if self.pid == event_obj['pid']:
                return
            if event_obj['comm'] == 'trace_event':
                return
        if event in self.event_handlers:
            for handler in self.event_handlers[event]:
                handler(event_obj)

    def start(self):
        self.build_trace_objs()
        time.sleep(1)
        cmd = [self.trace_event_elf]
        for obj in self.trace_bpf_o:
            cmd.append("-o")
            cmd.append(obj)
        for pid in self.pid_filter:
            cmd.append("-p")
            cmd.append(str(pid))
        for uid in self.uid_filter:
            cmd.append("-u")
            cmd.append(str(uid))
        cmd_str = " ".join(cmd)
        self.proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        json_str = None
        while True:
            line = self.proc.stdout.readline()
            if not line:
                break
            try:
                line = line.decode()
            except:
                continue
            if json_str == None:
                if line == "{\n":
                    json_str = line
            elif line == "}\n":
                json_str += line
                event_obj = json.loads(json_str)
                self.call_event_handlers(event_obj)
                json_str = None
            else:
                json_str += line
        self.proc = None

    def stop(self):
        if self.proc != None:
            self.proc.terminate()


